import base64
import logging
import pickle

import gevent
from PyTango.gevent import DeviceProxy

from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import (
    SampleChanger,
    SampleChangerState,
)
from mxcubecore.HardwareObjects.abstract.sample_changer.Container import (
    Container,
    Sample,
)
from mxcubecore.TaskUtils import task


class Pin(Sample):
    def __init__(self, basket, cell_no, basket_no, sample_no):
        super(Pin, self).__init__(
            basket, Pin.get_sample_address(cell_no, basket_no, sample_no), True
        )
        self._set_holder_length(22.0)
        self.present = True

    def get_basket_no(self):
        return self.get_container().get_index() + 1

    def get_vial_no(self):
        return self.get_index() + 1

    def get_cell_no(self):
        return self.get_container().get_container().get_index() + 1

    def get_cell(self):
        return self.get_container().get_container()

    @staticmethod
    def get_sample_address(cell_number, basket_number, sample_number):
        return (
            str(cell_number) + ":" + str(basket_number) + ":" + "%02d" % sample_number
        )


class Basket(Container):
    __TYPE__ = "Puck"

    def __init__(self, container, cell_no, basket_no, unipuck=False):
        super(Basket, self).__init__(
            self.__TYPE__,
            container,
            Basket.get_basket_address(cell_no, basket_no),
            True,
        )
        for i in range(16 if unipuck else 10):
            slot = Pin(self, cell_no, basket_no, i + 1)
            self._add_component(slot)
        self.present = True

    @staticmethod
    def get_basket_address(cell_number, basket_number):
        return str(cell_number) + ":" + str(basket_number)

    def get_cell_no(self):
        return self.get_container().get_index() + 1

    def get_cell(self):
        return self.get_container()

    def clear_info(self):
        self.get_container()._reset_basket_info(self.get_index() + 1)
        self.get_container()._trigger_info_changed_event()


class Cell(Container):
    __TYPE__ = "Cell"

    def __init__(self, container, number, sc3_pucks=True):
        super(Cell, self).__init__(
            self.__TYPE__, container, Cell.get_cell_address(number), True
        )
        self.present = True
        if sc3_pucks:
            for i in range(3):
                self._add_component(
                    Basket(self, number, i + 1, unipuck=1 - (number % 2))
                )
        else:
            for i in range(3):
                self._add_component(Basket(self, number, i + 1, unipuck=True))

    @staticmethod
    def get_cell_address(cell_number):
        return str(cell_number)

    def _reset_basket_info(self, basket_no):
        pass

    def clear_info(self):
        self.get_container()._reset_cell_info(self.get_index() + 1)
        self.get_container()._trigger_info_changed_event()

    def get_cell(self):
        return self


class FlexHCD(SampleChanger):
    __TYPE__ = "HCD"

    def __init__(self, *args, **kwargs):
        super(FlexHCD, self).__init__(self.__TYPE__, True, *args, **kwargs)

    def init(self):
        sc3_pucks = self.get_property("sc3_pucks", True)

        for i in range(8):
            cell = Cell(self, i + 1, sc3_pucks)
            self._add_component(cell)

        self.robot = self.get_property("tango_device")
        if self.robot:
            self.robot = DeviceProxy(self.robot)

        self.exporter_addr = self.get_property("exporter_address")

        if self.exporter_addr:
            self.swstate_attr = self.add_channel(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "swstate",
                },
                "State",
            )

        self.controller = self.get_object_by_role("controller")
        self.prepareLoad = self.get_command_object("moveToLoadingPosition")
        self.timeout = 3
        self.gripper_types = {
            -1: "No Gripper",
            1: "UNIPUCK",
            2: "MINISPINE",
            3: "FLIPPING",
            4: "UNIPUCK_DOUBLE",
            5: "PLATE",
        }

        SampleChanger.init(self)
        # self._set_state(SampleChangerState.Disabled)
        self._update_selection()
        self.state = self._read_state()

    @task
    def prepare_load(self):
        if self.controller:
            self.controller.hutch_actions(enter=True)
        else:
            self.prepareLoad()

    @task
    def _prepare_centring_task(self):
        if self.controller:
            self.controller.hutch_actions(enter=False, sc_loading=True)
        else:
            gevent.sleep(2)
            self.get_command_object("unlockMinidiffMotors")(wait=True)
            self.get_command_object("prepareCentring")(wait=True)

    def prepare_centring(self):
        self._prepare_centring_task()

    def get_sample_properties(self):
        return (Pin.__HOLDER_LENGTH_PROPERTY__,)

    def get_basket_list(self):
        basket_list = []
        # put here only the baskets that exist, not all the possible ones
        # if self.exporter_addr:
        #    basket_list =
        for cell in self.get_components():
            for basket in cell.get_components():
                if isinstance(basket, Basket):
                    basket_list.append(basket)

        return basket_list

    def _do_change_mode(self, *args, **kwargs):
        return

    def _do_update_info(self):
        # self._update_selection()
        self._update_state()

    def _do_scan(self, component, recursive=True, saved={"barcodes": None}):
        return

    def _execute_cmd(self, cmd, *args, **kwargs):
        timeout = kwargs.pop("timeout", None)
        if args:
            cmd_str = "flex.%s(%s)" % (cmd, ",".join(map(repr, args)))
        else:
            cmd_str = "flex.%s()" % cmd
        cmd_id = self.robot.eval(cmd_str)

        if not cmd_id:
            cmd_id = self.robot.eval(cmd_str)
        with gevent.Timeout(
            timeout, RuntimeError("Timeout while executing %s" % repr(cmd_str))
        ):
            while True:
                if self.robot.is_finished(cmd_id):
                    break
                gevent.sleep(0.2)

        res = self.robot.get_result(cmd_id)
        if res:
            res = pickle.loads(base64.b64decode(res))
            if isinstance(res, Exception):
                raise res
            else:
                return res

    def _execute_cmd_exporter(self, cmd, *args, **kwargs):
        ret = None
        timeout = kwargs.pop("timeout", 900)
        if args:
            args_str = "%s" % "\t".join(map(repr, args))
        if kwargs.pop("command", None):
            exp_cmd = self.add_command(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd,
            )
            if args:
                ret = exp_cmd(args_str)
            else:
                ret = exp_cmd()
        if kwargs.pop("attribute", None):
            exp_attr = self.add_channel(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd[3:],
            )
            if cmd.startswith("get"):
                return exp_attr.get_value()
            if cmd.startswith("set"):
                ret = exp_attr.set_value(args_str)

        self._wait_ready(timeout=timeout)
        return ret

    def _assert_ready(self):
        if self.exporter_addr:
            if not self._ready():
                raise RuntimeError("Sample changer is busy cant mount/unmount")

    def _ready(self):
        return self.swstate_attr.get_value() == "Ready"

    def _wait_ready(self, timeout=None):
        err_msg = "Timeout waiting for sample changer to be ready"
        # None means infinite timeout <=0 means default timeout
        if timeout is not None and timeout <= 0:
            timeout = self.timeout
        with gevent.Timeout(timeout, RuntimeError(err_msg)):
            while not self._ready():
                gevent.sleep(0.5)

    def _do_select(self, component):
        if isinstance(component, Cell):
            cell_pos = component.get_index() + 1
        elif isinstance(component, Basket) or isinstance(component, Pin):
            cell_pos = component.get_cell_no()

        if self.exporter_addr:
            self._execute_cmd_exporter("moveDewar", cell_pos, command=True)
        else:
            self._execute_cmd("moveDewar", cell_pos)

        self._update_selection()

    def chained_load(self, old_sample, sample):
        self._assert_ready()
        if self.exporter_addr:
            unload_load_task = gevent.spawn(
                self._execute_cmd_exporter,
                "loadSample",
                sample.get_cell_no(),
                sample.get_basket_no(),
                sample.get_vial_no(),
                command=True,
            )
        else:
            unload_load_task = gevent.spawn(
                self._execute_cmd,
                "chainedUnldLd",
                [
                    old_sample.get_cell_no(),
                    old_sample.get_basket_no(),
                    old_sample.get_vial_no(),
                ],
                [sample.get_cell_no(), sample.get_basket_no(), sample.get_vial_no()],
            )

        gevent.sleep(10)

        err_msg = "Timeout waiting for sample changer to be in safe position"
        while not unload_load_task.ready():
            if self.exporter_addr:
                loading_state = self._execute_cmd_exporter(
                    "getCurrentLoadSampleState", attribute=True
                )
                if "on_gonio" in loading_state:
                    self._set_loaded_sample(sample)
                    with gevent.Timeout(60, RuntimeError(err_msg)):
                        logging.getLogger("HWR").info(err_msg)
                        while not self._execute_cmd_exporter(
                            "getRobotIsSafe", attribute=True
                        ):
                            gevent.sleep(0.5)
                    return True
            else:
                loading_state = str(
                    self._execute_cmd("get_robot_cache_variable", "LoadSampleStatus")
                )
                if "on_gonio" in loading_state:
                    self._set_loaded_sample(sample)
                    with gevent.Timeout(60, RuntimeError(err_msg)):
                        logging.getLogger("HWR").info(err_msg)
                        while (
                            not self._execute_cmd(
                                "get_robot_cache_variable", "data:dioRobotIsSafe"
                            )
                            == "true"
                        ):
                            gevent.sleep(0.5)
                    return True
            gevent.sleep(2)

        logging.getLogger("HWR").info("unload load task done")
        for msg in self.get_robot_exceptions():
            logging.getLogger("HWR").error(msg)

        return self._check_pin_on_gonio()

    def _check_pin_on_gonio(self):
        if self.exporter_addr:
            _on_gonio = self._execute_cmd_exporter("pin_on_gonio", command=True)
        else:
            _on_gonio = self._execute_cmd("pin_on_gonio")

        if _on_gonio:
            # finish the loading actions
            self._prepare_centring_task()
            return True
        else:
            logging.getLogger("HWR").info("reset loaded sample")
            self._reset_loaded_sample()
            # if self.controller:
            #    self.controller.hutch_actions(release_interlock=True)
            return False

    def reset_loaded_sample(self):
        if self.exporter_addr:
            self._execute_cmd_exporter("resetLoadedPosition", command=True)
        else:
            self._execute_cmd("reset_loaded_position")
        self._reset_loaded_sample()

    def get_robot_exceptions(self):
        if self.exporter_addr:
            """
            return self._execute_cmd_exporter('getRobotExceptions',
                                              attribute=True)
            """
            return ""
        else:
            return self._execute_cmd("getRobotExceptions")

    @task
    def load(self, sample):
        self.prepare_load(wait=True)
        self.enable_power()
        try:
            res = SampleChanger.load(self, sample)
        finally:
            for msg in self.get_robot_exceptions():
                logging.getLogger("HWR").error(msg)
        if res:
            self.prepare_centring()
        return res

    @task
    def unload_sample(
        self,
        holderLength,
        sample_id=None,
        sample_location=None,
        successCallback=None,
        failureCallback=None,
    ):
        self._assert_ready()
        cell, basket, sample = sample_location
        sample = self.get_component_by_address(
            Pin.get_sample_address(cell, basket, sample)
        )
        return self.unload(sample)

    @task
    def unload(self, sample):
        self.prepare_load(wait=True)
        self.enable_power()

        if not sample:
            sample = self.get_loaded_sample().get_address()

        try:
            SampleChanger.unload(self, sample)
        finally:
            for msg in self.get_robot_exceptions():
                logging.getLogger("HWR").error(msg)

    def get_gripper(self):
        if self.exporter_addr:
            gripper_type = self._execute_cmd_exporter(
                "get_gripper_type", attribute=True
            )
        else:
            gripper_type = self._execute_cmd("get_gripper_type")

        return self.gripper_types.get(gripper_type, "?")

    def get_available_grippers(self):
        grippers = []

        try:
            if self.exporter_addr:
                ret = sorted(
                    self._execute_cmd_exporter("getSupportedGrippers", attribute=True)
                )
                for gripper in ret:
                    grippers.append(self.gripper_types[gripper])
            else:
                ret = [1, 3]  # self._execute_cmd("get_supported_grippers")

                for gripper in ret:
                    grippers.append(self.gripper_types[gripper])
        except Exception:
            grippers = [-1]

        return grippers

    @task
    def change_gripper(self, gripper=None):
        self.prepare_load(wait=True)
        self.enable_power()
        if self.exporter_addr:
            if gripper:
                self._execute_cmd_exporter("setGripper", gripper, command=True)
            else:
                self._execute_cmd_exporter("changeGripper", command=True)
        else:
            self._execute_cmd("changeGripper")

    @task
    def home(self):
        self.prepare_load(wait=True)
        self.enable_power()
        if self.exporter_addr:
            self._execute_cmd_exporter("homeClear", command=True)
        else:
            self._execute_cmd("homeClear")

    @task
    def enable_power(self):
        if not self.exporter_addr:
            self._execute_cmd("enablePower", 1)

    @task
    def defreeze(self):
        self.prepare_load(wait=True)
        self.enable_power()
        if self.exporter_addr:
            self._execute_cmd_exporter("defreezeGripper", command=True)
        else:
            self._execute_cmd("defreezeGripper")

    def _do_load(self, sample=None):
        self._update_state()

        if self.exporter_addr:
            load_task = gevent.spawn(
                self._execute_cmd_exporter,
                "loadSample",
                sample.get_cell_no(),
                sample.get_basket_no(),
                sample.get_vial_no(),
                command=True,
            )
        else:
            load_task = gevent.spawn(
                self._execute_cmd,
                "loadSample",
                sample.get_cell_no(),
                sample.get_basket_no(),
                sample.get_vial_no(),
            )

        gevent.sleep(10)
        err_msg = "Timeout waiting for sample changer to be in safe position"
        while not load_task.ready():
            if self.exporter_addr:
                loading_state = self._execute_cmd_exporter(
                    "getCurrentLoadSampleState", attribute=True
                )

                if "on_gonio" in loading_state:
                    self._set_loaded_sample(sample)
                    with gevent.Timeout(20, RuntimeError(err_msg)):
                        while not self._execute_cmd_exporter(
                            "getRobotIsSafe", attribute=True
                        ):
                            gevent.sleep(0.5)
                    return True
            else:
                loading_state = str(
                    self._execute_cmd("get_robot_cache_variable", "LoadSampleStatus")
                )
                if "on_gonio" in loading_state:
                    self._set_loaded_sample(sample)
                    with gevent.Timeout(20, RuntimeError(err_msg)):
                        while (
                            not self._execute_cmd(
                                "get_robot_cache_variable", "data:dioRobotIsSafe"
                            )
                            == "true"
                        ):
                            gevent.sleep(0.5)
                    return True
            gevent.sleep(2)

        if self.exporter_addr:
            loaded_sample = self._execute_cmd_exporter(
                "get_loaded_sample", attribute=True
            )
        else:
            loaded_sample = self._execute_cmd("get_loaded_sample")
        if loaded_sample == (
            sample.get_cell_no(),
            sample.get_basket_no(),
            sample.get_vial_no(),
        ):
            self._set_loaded_sample(sample)
            return True
        return self._check_pin_on_gonio()

    def _do_unload(self, sample=None):
        loaded_sample = self.get_loaded_sample()
        if loaded_sample is not None and loaded_sample != sample:
            raise RuntimeError("Cannot unload another sample")

        if self.exporter_addr:
            self._execute_cmd_exporter(
                "unloadSample",
                sample.get_cell_no(),
                sample.get_basket_no(),
                sample.get_vial_no(),
                command=True,
            )
            loaded_sample = self._execute_cmd_exporter(
                "getLoadedSample", attribute=True
            )
        else:
            self._execute_cmd(
                "unloadSample",
                sample.get_cell_no(),
                sample.get_basket_no(),
                sample.get_vial_no(),
            )
            loaded_sample = self._execute_cmd("get_loaded_sample")
        if loaded_sample == (-1, -1, -1):
            self._reset_loaded_sample()
            if self.controller:
                self.controller.hutch_actions(release_interlock=True)
            return True

        return False

    def _do_abort(self):
        if self.exporter_addr:
            self._execute_cmd_exporter("abort", command=True)
        else:
            self._execute_cmd("abort")

    def _do_reset(self):
        if self.controller:
            self.controller.hutch_actions(enter=True)
        if self.exporter_addr:
            self._execute_cmd_exporter("homeClear", command=True)
        else:
            self._execute_cmd("homeClear")

    def clear_basket_info(self, basket):
        return self._reset_basket_info(basket)

    def _reset_basket_info(self, basket):
        pass

    def clear_cell_info(self, cell):
        return self._reset_cell_info(cell)

    def _reset_cell_info(self, cell):
        pass

    def _update_state(self):
        # see if the command exists for exporter
        if not self.exporter_addr:
            pass
            # defreezing = self._execute_cmd("isDefreezing")

            # if defreezing:
            #    self._set_state(SampleChangerState.Moving)

        try:
            state = self._read_state()
        except Exception:
            state = SampleChangerState.Unknown

        self._set_state(state)

    def is_sequencer_ready(self):
        if self.prepareLoad:
            cmdobj = self.get_command_object
            return all(
                [cmd.isSpecReady() for cmd in (cmdobj("moveToLoadingPosition"),)]
            )
        return True

    def _read_state(self):
        # should read state from robot
        if self.exporter_addr:
            state = self.swstate_attr.get_value().upper()
        else:
            state = "RUNNING" if self._execute_cmd("robot.isBusy") else "STANDBY"
            if state == "STANDBY" and not self.is_sequencer_ready():
                state = "RUNNING"

        state_converter = {
            "ALARM": SampleChangerState.Alarm,
            "FAULT": SampleChangerState.Fault,
            "RUNNING": SampleChangerState.Moving,
            "READY": SampleChangerState.Ready,
            "STANDBY": SampleChangerState.Ready,
        }

        return state_converter.get(state, SampleChangerState.Unknown)

    def _is_device_busy(self, state=None):
        if state is None:
            state = self._read_state()
        return state not in (
            SampleChangerState.Ready,
            SampleChangerState.Loaded,
            SampleChangerState.Alarm,
            SampleChangerState.Disabled,
            SampleChangerState.Fault,
            SampleChangerState.StandBy,
        )

    def _is_device_ready(self):
        state = self._read_state()
        return state in (SampleChangerState.Ready, SampleChangerState.Charging)

    def _wait_device_ready(self, timeout=None):
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self._is_device_ready():
                gevent.sleep(0.01)

    def _update_selection(self):
        if self.exporter_addr:
            sample_cell, sample_puck, sample = self._execute_cmd_exporter(
                "get_loaded_sample", attribute=True
            )
            cell = sample_cell
            puck = sample_puck
        else:
            cell, puck = self._execute_cmd("get_cell_position")
            sample_cell, sample_puck, sample = self._execute_cmd("get_loaded_sample")

        for c in self.get_components():
            i = c.get_index()
            if cell == i + 1:
                self._set_selected_component(c)
                break

        # find sample
        for s in self.get_sample_list():
            if s.get_coords() == (sample_cell, sample_puck, sample):
                self._set_loaded_sample(s)
                self._set_selected_sample(s)
                return

        for s in self.get_sample_list():
            s._set_loaded(False)
        self._set_selected_sample(None)

    def prepare_hutch(self, **kwargs):
        if self.exporter_addr:
            return

        user_port = kwargs.get("user_port")
        robot_port = kwargs.get("robot_port")
        if user_port is not None:
            self._execute_cmd("robot.user_port(user_port)")

        if robot_port is not None:
            self._execute_cmd("robot.robot_port(robot_port)")



"""Custom MicroMAX Beamline object.

Adds support for `sample_delivery` configurable, which
specifies the sample delivery mode for MXCuBE.

Following sample delivery modes are supported:

* osc - Oscillation sample delivery
* hve - HVE (injector) sample delivery

Example of `sample_delivery` configuration::

  configuration:
    sample_delivery: osc
"""

from enum import Enum

import mxcubecore.HardwareObjects.MAXIV.beamline


class SampleDelivery(Enum):
    osc = "osc"
    hve = "hve"


class Beamline(mxcubecore.HardwareObjects.MAXIV.beamline.Beamline):
    def __init__(self, name):
        super().__init__(name)

        # 'cached' Sample Delivery mode config
        self._sample_delivery = None

    #
    # 'sample_delivery' config
    #

    def _load_sample_delivery(self):
        val = self.get_property("sample_delivery", SampleDelivery.osc.value)
        self._sample_delivery = SampleDelivery(val)

    @property
    def sample_delivery(self) -> SampleDelivery:
        if self._sample_delivery is None:
            self._load_sample_delivery()

        return self._sample_delivery

    def is_hve_sample_delivery(self) -> bool:
        """True when HVE sample delivery mode is configured."""
        return self.sample_delivery == SampleDelivery.hve



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
BeamDefinerMockup class.
Two mock motors change the width and the height of the beam.
Example xml configuration:

.. code-block:: xml

 <object class="BeamDefinerMockup">
   <username>Beam Definer</username>
   <object hwrid="/beam_size_ver" role="beam_size_ver" />
   <object hwrid="/beam_size_hor" role="beam_size_hor" />
   <values>{"50x50": (0.05, 0.05), "100x100": (0.1, 0.1), "20x5": (0.02, 0.005)}</values>
   <default_size_name>50x50</default_size_name>
 </object>
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"

import random
import time

from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState
from mxcubecore.HardwareObjects.mockup.ActuatorMockup import ActuatorMockup


class BeamDefinerMockup(AbstractNState, ActuatorMockup):
    """BeamDefinerMockup class"""

    def __init__(self, *args):
        super().__init__(*args)
        self.beam_size_hor = None
        self.beam_size_ver = None

    def init(self):
        super().init()
        self.beam_size_hor = self.get_object_by_role("beam_size_hor")
        self.beam_size_ver = self.get_object_by_role("beam_size_ver")

        # set default value [mm]
        self.beam_size_hor.set_value(0.1)
        self.beam_size_ver.set_value(0.1)

        _default_name = self.get_property("default_size_name")
        if self.get_value() == self.VALUES.UNKNOWN and _default_name:
            # set default beam value
            self._set_value(self.VALUES[_default_name])
        self.update_value()

        self.connect(self.beam_size_hor, "valueChanged", self.motors_changed)
        self.connect(self.beam_size_ver, "valueChanged", self.motors_changed)

    def motors_changed(self, value):
        """Emit valueChanged for the definer when motor position changed"""
        print(value)
        name = self.get_current_position_name()
        self.emit("valueChanged", name)

    def get_value(self):
        """Get the beam value.
        Returns:
            (Enum): The current position Enum.
        """
        try:
            return self.VALUES[self.get_current_position_name()]
        except (ValueError, KeyError, TypeError):
            return self.VALUES.UNKNOWN

    def get_current_position_name(self):
        """Get the current beam size name.
        Returns:
            (str): Current beam size name.
        """
        hor = self.beam_size_hor.get_value()
        ver = self.beam_size_ver.get_value()
        for val in self.VALUES:
            if val.value == (hor, ver):
                return val.name
        return "UNKNOWN"

    def _set_value(self, value):
        """Set the beam size.
        Args:
            value(str): name of the beam size to set.
        """
        if isinstance(value, str):
            size_x, size_y = self.VALUES[value].value
        else:
            size_x, size_y = value.value
        self.beam_size_hor.set_value(float(size_x))
        self.beam_size_ver.set_value(float(size_y))
        time.sleep(random.uniform(0.3, 1.0))
        self.update_value(value)

    def get_predefined_positions_list(self):
        """Get the position labels list.
        Returns:
            (list): List of all the labels defined.
        """
        values = []
        for value in self.VALUES:
            nam = value.name
            if value.name != "UNKNOWN":
                values.append(nam)
        return values



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Example xml file:
<object class="ExporterMotor">
  <username>phiy</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <actuator_name>AlignmentY</actuator_name>
  <tolerance>1e-2</tolerance>
</object>
optional parameters, with default values. To be used only if the values
differ from the default ones.
  The position attribute is actuator_name+position_suffix
  <position_suffix>Position</position_suffix>
  The state attribute is actuator_name+state_suffix
  <state_suffix>State</state_suffix>
  Use the global application state instead of the motor state.
  <use_global_state>False</use_global_state>
"""

import logging
import math
import sys

from gevent import (
    Timeout,
    sleep,
)

from mxcubecore.Command.Exporter import Exporter
from mxcubecore.Command.exporter.ExporterStates import ExporterStates
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class ExporterMotor(AbstractMotor):
    """Motor using the Exporter protocol, based on AbstractMotor"""

    def __init__(self, name):
        super().__init__(name)
        self._motor_pos_suffix = None
        self._motor_state_suffix = None
        self._exporter = None
        self._exporter_address = None
        self.motor_position_chan = None
        self.motor_state_chan = None
        self.use_state = None

    def init(self):
        """Initialise the motor"""
        super().init()

        self._motor_pos_suffix = self.get_property("position_suffix", "Position")
        self._motor_state_suffix = self.get_property("state_suffix", "State")
        self.use_state = self.get_property("use_global_state", False)

        self._exporter_address = self.get_property("exporter_address")
        _host, _port = self._exporter_address.split(":")
        self._exporter = Exporter(_host, int(_port))

        self.motor_position_chan = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self._exporter_address,
                "name": "position",
            },
            self.actuator_name + self._motor_pos_suffix,
        )
        if self.motor_position_chan:
            self.get_value()
            self.motor_position_chan.connect_signal("update", self.update_value)

        if self.use_state:
            _name = "State"
        else:
            _name = self.actuator_name + self._motor_state_suffix
        self.motor_state_chan = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self._exporter_address,
                "name": "motor_state",
            },
            _name,
        )

        if self.motor_state_chan:
            self.motor_state_chan.connect_signal("update", self._update_state)

        self.update_state()

    def get_state(self):
        """Get the motor state.
        Returns:
            (enum 'HardwareObjectState'): Motor state.
        """
        try:
            _state = self.motor_state_chan.get_value().upper()
            self.specific_state = _state
            return ExporterStates[_state].value
        except (KeyError, AttributeError):
            return self.STATES.UNKNOWN

    def _update_state(self, state):
        try:
            state = state.upper()
            state = ExporterStates[state].value
        except (AttributeError, KeyError):
            state = self.STATES.UNKNOWN
        return self.update_state(state)

    def _get_hwstate(self):
        """Get the hardware state, reported by the MD2 application.
        Returns:
            (string): The state.
        """
        try:
            return self._exporter.read_property("HardwareState")
        except Exception:
            return "Ready"

    def _get_swstate(self):
        """Get the software state, reported by the MD2 application.
        Returns:
            (string): The state.
        """
        return self._exporter.read_property("State")

    def _ready(self):
        """Get the "Ready" state - software and hardware.
        Returns:
            (bool): True if both "Ready", False otherwise.
        """
        if (
            self._get_swstate() == "Ready"
            and self._get_hwstate() == "Ready"
            and self.motor_state_chan.get_value() == "Ready"
        ):
            return True
        return False

    def _wait_ready(self, timeout=3):
        """Wait for the state to be "Ready".
        Args:
            timeout (float): waiting time [s].
        Raises:
            RuntimeError: Execution timeout.
        """
        with Timeout(timeout, RuntimeError("Execution timeout")):
            while not self._ready():
                sleep(0.01)

    def wait_move(self, timeout=20):
        """Wait until the end of move ended, using the application state.
        Args:
            timeout(float): Timeout [s]. Default value is 20 s
        """
        self._wait_ready(timeout)

    def wait_motor_move(self, timeout=20):
        """Wait until the end of move ended using the motor state.
        Args:
            timeout(float): Timeout [s]. Default value is 20 s
        Raises:
            RuntimeError: Execution timeout.
        """
        with Timeout(timeout, RuntimeError("Execution timeout")):
            while self.get_state() != self.STATES.READY:
                sleep(0.01)

    def get_value(self):
        """Get the motor position.
        Returns:
            (float): Motor position.
        """
        _val = self.motor_position_chan.get_value()

        if _val is None or math.isnan(_val):
            logging.getLogger("HWR").debug("Value of %s is NaN" % self.actuator_name)
            _val = self._nominal_value

        self._nominal_value = _val

        return self._nominal_value

    def __get_limits(self, cmd):
        """Returns motor low and high limits.
        Args:
            cmd (str): command name
        Returns:
            (tuple): two floats tuple (low limit, high limit).
        """
        try:
            _low, _high = self._exporter.execute(cmd, (self.actuator_name,))
            # inf is a problematic value, convert to sys.float_info.max
            if _low == float("-inf"):
                _low = -sys.float_info.max

            if _high == float("inf"):
                _high = sys.float_info.max

            return _low, _high
        except Exception:
            return self._nominal_limits

    def get_limits(self):
        """Returns motor low and high limits.
        Args:
            cmd (str): command name
        Returns:
            (tuple): two floats tuple (low limit, high limit).
        """
        self._nominal_limits = self.__get_limits("getMotorLimits")
        return self._nominal_limits

    def get_dynamic_limits(self):
        """Returns motor low and high dynamic limits.
        Returns:
            (tuple): two floats tuple (low limit, high limit).
        """
        return self.__get_limits("getMotorDynamicLimits")

    def _set_value(self, value):
        """Move motor to absolute value.
        Args:
            value (float): target value
        """
        self.update_state(self.STATES.BUSY)
        self.motor_position_chan.set_value(value)

    def abort(self):
        """Stop the motor movement immediately."""
        if self.get_state() != self.STATES.UNKNOWN:
            self._exporter.execute("abort")

    def home(self, timeout=None):
        """Homing procedure.
        Args:
            timeout (float): optional - timeout [s].
        """
        self._exporter.execute("startHomingMotor", (self.actuator_name,))
        self.wait_ready(timeout)

    def get_max_speed(self):
        """Get the motor maximum speed.
        Returns:
            (float): the maximum speed [unit/s].
        """
        return self._exporter.execute("getMotorMaxSpeed", (self.actuator_name,))



# -*- coding: utf-8 -*-
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract Energy and Wavelength class.
Defines the get/set wavelength, get_wavelength_limits methods and is_tunable
property. Implements update_value.
Emits signals valueChanged and attributeChanged.
"""

import abc

from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator
from mxcubecore.utils.conversion import HC_OVER_E

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractEnergy(AbstractActuator):
    """Abstract Energy and Wavelength"""

    # Unit for 'value' attribute
    unit = "keV"

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self._wavelength_limits = (None, None)

    def is_ready(self):
        """Check if the state is ready.
        Returns:
            (bool): True if ready, False otherwise.
        """
        if self.read_only:
            return True
        return super().is_ready()

    @property
    def is_tunable(self):
        """Check if not fixed energy.
        Returns:
            (bool): True if tunable, False if fixed energy.
        """
        return not self.read_only

    def get_wavelength(self):
        """Read the wavelength
        Returns:
            (float): Wavelength [Å].
        """
        return self.calculate_wavelength()

    def get_wavelength_limits(self):
        """Return wavelength low and high limits.
        Returns:
            (tuple): two floats tuple (low limit, high limit) [Å].
        """
        _low, _high = self.get_limits()
        self._wavelength_limits = (
            self.calculate_wavelength(_high),
            self.calculate_wavelength(_low),
        )
        return self._wavelength_limits

    def set_wavelength(self, value, timeout=None):
        """Move motor to absolute value. Wait the move to finish.
        Args:
            value (float): target position [keV]
            timeout (float): optional - timeout [s].
                             if timeout = 0: return at once and do not wait
                             if timeout is None: wait forever
        """
        self.set_value(self.calculate_energy(value), timeout=timeout)

    def calculate_wavelength(self, energy=None):
        """Calculate wavelength from energy
        Args:
            energy(float): Energy [keV]
        Returns:
            (float): wavelength [Å]
        """
        energy = energy or self.get_value()

        # TODO NBNB This is naughty. Could  we not put the heuristic switch
        #  in the calling functions, to avoid surprises?
        #  rhfogh 20210826
        # energy in KeV to get wavelength in Å
        energy = energy / 1000.0 if energy > 1000 else energy

        return HC_OVER_E / energy

    def calculate_energy(self, wavelength=None):
        """Calculate energy from wavelength
        Args:
            value((float): wavelength [Å]
        Returns:
            (float): Energy [keV]
        """
        wavelength = wavelength or self.get_wavelength()
        return HC_OVER_E / wavelength

    def update_value(self, value=None):
        """Emist signal energyChanged for both energy and wavelength
        Argin:
            value: Not used, but kept in the method signature.
        """

        if value is None:
            value = self.get_value()
        self._nominal_value = value

        _wavelength_value = self.calculate_wavelength(value)
        self.emit("energyChanged", (value, _wavelength_value))
        self.emit("valueChanged", (value,))

    def force_emit_signals(self):
        super().force_emit_signals()
        _energy_value = self.get_value()
        _wavelength_value = self.calculate_wavelength(_energy_value)
        self.emit("energyChanged", (_energy_value, _wavelength_value))



"""Tango Shutter Hardware Object
Example XML::

  <object class="ALBAEpsActuator">
    <username>Photon Shutter</username>
    <taurusname>bl13/ct/eps-plc-01</taurusname>
    <channel type="sardana" polling="events" name="actuator">pshu</channel>
    <states>Open,Closed</states>
  </object>


Public Interface:
   Commands:
       int get_state()
           Description:
               returns current state
           Output:
               integer value describing the state
               current states correspond to:
                      0: out
                      1: in
                      9: moving
                     11: alarm
                     13: unknown
                     23: fault

       string getStatus()
           Description:
               returns current state as a string that can contain a more
               descriptive information about current state

           Output:
               status string

       cmdIn()
           Executes the command associated to the "In" action
       cmdOut()
           Executes the command associated to the "Out" action

   Signals:
       stateChanged

"""

import logging

from mxcubecore import BaseHardwareObjects

STATE_OUT, STATE_IN, STATE_MOVING, STATE_FAULT, STATE_ALARM, STATE_UNKNOWN = (
    0,
    1,
    9,
    11,
    13,
    23,
)


class ALBAFastShutter(BaseHardwareObjects.HardwareObject):
    states = {
        STATE_OUT: "out",
        STATE_IN: "in",
        STATE_MOVING: "moving",
        STATE_FAULT: "fault",
        STATE_ALARM: "alarm",
        STATE_UNKNOWN: "unknown",
    }

    default_state_strings = ["Out", "In"]

    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.actuator_state = STATE_UNKNOWN
        self.actuator_value = None
        self.motor_position = None
        self.motor_state = None

        try:
            self.nistart_cmd = self.get_command_object("nistart")
            self.nistop_cmd = self.get_command_object("nistop")

            self.actuator_channel = self.get_channel_object("actuator")
            self.motorpos_channel = self.get_channel_object("motorposition")
            self.motorstate_channel = self.get_channel_object("motorstate")

            self.actuator_channel.connect_signal("update", self.stateChanged)
            self.motorpos_channel.connect_signal("update", self.motor_positions_changed)
            self.motorstate_channel.connect_signal("update", self.motorStateChanged)
        except KeyError:
            logging.getLogger().warning("%s: cannot report FrontEnd State", self.name())

        try:
            state_string = self.get_property("states")
            if state_string is None:
                self.state_strings = self.default_state_strings
            else:
                states = state_string.split(",")
                self.state_strings = states[1].strip(), states[0].strip()
        except Exception:
            import traceback

            logging.getLogger("HWR").warning(traceback.format_exc())
            self.state_strings = self.default_state_strings

    def get_state(self):
        if self.actuator_state == STATE_UNKNOWN:
            self.actuator_value = self.actuator_channel.get_value()
            self.motor_position = self.motorpos_channel.get_value()
            self.motor_state = self.motorstate_channel.get_value()
            self.update_state()
        return self.actuator_state

    def update_state(self):
        if None in [self.actuator_value, self.motor_position, self.motor_state]:
            act_state = STATE_UNKNOWN
        elif str(self.motor_state) == "MOVING":
            act_state = STATE_MOVING
        elif str(self.motor_state) != "ON" or abs(self.motor_position) > 0.01:
            act_state = STATE_ALARM
        else:
            state = self.actuator_value.lower()

            if state == "high":
                act_state = STATE_OUT
            else:
                act_state = STATE_IN

        if act_state != self.actuator_state:
            self.actuator_state = act_state
            self.emitStateChanged()

    def stateChanged(self, value):
        self.actuator_value = value
        self.update_state()

    def motor_positions_changed(self, value):
        self.motor_position = value
        self.update_state()

    def motorStateChanged(self, value):
        self.motor_state = value
        self.update_state()

    def emitStateChanged(self):
        #
        # emit signal
        #
        self.emit("fastStateChanged", ((self.actuator_state),))

    def getMotorPosition(self):
        if self.motor_position is None:
            self.motor_position = self.motorpos_channel.get_value()
        return self.motor_position

    def getMotorState(self):
        if self.motor_state is None:
            self.motor_state = self.motorstate_channel.get_value()
        return self.motor_state

    def getUserName(self):
        return self.username

    def getStatus(self):
        """ """
        state = self.get_state()

        if state in [STATE_OUT, STATE_IN]:
            return self.state_strings[state]
        elif state in self.states:
            return self.states[state]
        else:
            return "Unknown"

    def cmdIn(self):
        self.open()

    def cmdOut(self):
        self.close()

    def close(self):
        self.motorpos_channel.set_value(0)
        self.set_ttl("High")

    def open(self):
        self.motorpos_channel.set_value(0)
        self.set_ttl("Low")

    def set_ttl(self, value):
        self.nistop_cmd()
        self.actuator_channel.set_value(value)
        self.nistart_cmd()

    def is_open(self):
        if self.actuator_state == STATE_IN:
            return True
        else:
            return False

    def is_close(self):
        if self.actuator_state == STATE_OUT:
            return True
        else:
            return False


def test_hwo(hwo):
    print("Name is: ", hwo.getUserName())

    print("Shutter state is: ", hwo.get_state())
    print("Shutter status is: ", hwo.getStatus())
    print("Motor position is: ", hwo.getMotorPosition())
    print("Motor state is: ", hwo.getMotorState())
    # hwo.open()
    # time.sleep(2)
    # print "is_open?" , hwo.is_open()
    # print "is_close?" , hwo.is_close()


if __name__ == "__main__":
    test()



import binascii
import logging
import os
import pprint
import socket
import time

import gevent
import requests

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

# import threading
from mxcubecore.HardwareObjects.SecureXMLRpcRequestHandler import (
    SecureXMLRpcRequestHandler,
)


class State(object):
    """
    Class for mimic the PyTango state object
    """

    def __init__(self, parent):
        self._value = "ON"
        self._parent = parent

    def get_value(self):
        return self._value

    def set_value(self, new_value):
        self._value = new_value
        self._parent.state_changed(new_value)

    def del_value(self):
        pass

    value = property(get_value, set_value, del_value, "Property for value")


class EdnaWorkflow(HardwareObject):
    """
    This HO acts as a interface to the Passerelle EDM workflow engine.

    The previous version of this HO was a Tango client. In order to avoid
    too many changes this version of the HO is a drop-in replacement of the
    previous version, hence the "State" object which mimics the PyTango state.

    Example of a corresponding XML file (currently called "ednaparams.xml"):

    <object class = "EdnaWorkflow" role = "workflow">
    <bes_host>mxhpc2-1705</bes_host>
    <bes_port>8090</bes_port>
    <object href="/session" role="session"/>
    <workflow>
        <title>MXPressA</title>
        <path>MXPressA</path>
    </workflow>
    <workflow>
        <title>...</title>
        <path>...</path>
    </workflow>
    """

    def __init__(self, name):
        HardwareObject.__init__(self, name)
        self._state = State(self)
        self.command_failed = False
        self.bes_workflow_id = None
        self.gevent_event = None
        self.bes_host = None
        self.bes_port = None
        self.token = None

    def _init(self):
        pass

    def init(self):
        self.gevent_event = gevent.event.Event()
        self.bes_host = self.get_property("bes_host")
        self.bes_port = int(self.get_property("bes_port"))
        self._state.value = "ON"

    def getState(self):
        return self._state

    def setState(self, new_state):
        self._state = new_state

    def delState(self):
        pass

    state = property(getState, setState, delState, "Property for state")

    def command_failure(self):
        return self.command_failed

    def set_command_failed(self, *args):
        logging.getLogger("HWR").error("Workflow '%s' Tango command failed!", args[1])
        self.command_failed = True

    def state_changed(self, new_value):
        new_value = str(new_value)
        logging.getLogger("HWR").debug(f"{self.name}: state changed to {new_value}")
        self.emit("stateChanged", (new_value,))

    def workflow_end(self):
        """
        The workflow has finished, sets the state to 'ON'
        """
        # If necessary unblock dialog
        if not self.gevent_event.is_set():
            self.gevent_event.set()
        self.state.value = "ON"

    def open_dialog(self, dict_dialog):
        # If necessary unblock dialog
        if not self.gevent_event.is_set():
            self.gevent_event.set()
        self.params_dict = {}
        if "reviewData" in dict_dialog and "inputMap" in dict_dialog:
            review_data = dict_dialog["reviewData"]
            for dict_entry in dict_dialog["inputMap"]:
                if "value" in dict_entry:
                    value = dict_entry["value"]
                else:
                    value = dict_entry["defaultValue"]
                self.params_dict[dict_entry["variableName"]] = str(value)
            self.emit("parametersNeeded", (review_data,))
            self.state.value = "OPEN"
            self.gevent_event.clear()
            while not self.gevent_event.is_set():
                self.gevent_event.wait()
                time.sleep(0.1)
        return self.params_dict

    def get_values_map(self):
        return self.params_dict

    def set_values_map(self, params):
        self.params_dict = params
        self.gevent_event.set()

    def get_available_workflows(self):
        workflow_list = []

        for _wf in self.get_property("workflow"):
            wf = dict(_wf)
            workflow_list.append(wf)
            wf["requires"] = [r.strip() for r in wf.get("requires", "").split(",")]
            wf["doc"] = ""

        return workflow_list

    def abort(self):
        self.generateNewToken()
        logging.getLogger("HWR").info("Aborting current workflow")
        # If necessary unblock dialog
        if not self.gevent_event.is_set():
            self.gevent_event.set()
        self.command_failed = False
        if self.bes_workflow_id is not None:
            abort_URL = (
                f"http://{self.bes_host}:{self.bes_port}/ABORT/{self.bes_workflow_id}"
            )
            logging.getLogger("HWR").info("BES abort web service URL: %r", abort_URL)
            response = requests.get(abort_URL)
            if response.status_code == 200:
                workflow_status = response.text
                logging.getLogger("HWR").info(
                    "BES workflow id %s: %s", self.bes_workflow_id, workflow_status
                )
        self.state.value = "ON"

    def generateNewToken(self):
        # See: https://wyattbaldwin.com/2014/01/09/generating-random-tokens-in-python/
        self.token = binascii.hexlify(os.urandom(5)).decode("utf-8")
        SecureXMLRpcRequestHandler.setReferenceToken(self.token)

    def getToken(self):
        return self.token

    def start(self, list_arguments):
        self.generateNewToken()
        # If necessary unblock dialog
        if not self.gevent_event.is_set():
            self.gevent_event.set()
        self.state.value = "RUNNING"

        self.dict_parameters = {}
        index = 0
        if len(list_arguments) == 0:
            self.error_stream("ERROR! No input arguments!")
            return
        elif len(list_arguments) % 2 != 0:
            self.error_stream("ERROR! Odd number of input arguments!")
            return
        while index < len(list_arguments):
            self.dict_parameters[list_arguments[index]] = list_arguments[index + 1]
            index += 2
        logging.info("Input arguments:")
        logging.info(pprint.pformat(self.dict_parameters))

        if "modelpath" in self.dict_parameters:
            modelpath = self.dict_parameters["modelpath"]
            if "." in modelpath:
                modelpath = modelpath.split(".")[0]
            self.workflow_name = os.path.basename(modelpath)
        else:
            self.error_stream("ERROR! No modelpath in input arguments!")
            return

        time0 = time.time()
        self.startBESWorkflow()
        time1 = time.time()
        logging.info("Time to start workflow: %f sec", time1 - time0)

    def startBESWorkflow(self):
        logging.info("Starting workflow %s", self.workflow_name)

        xml_rpc_server = HWR.beamline.xml_rpc_server
        if xml_rpc_server is None:
            logging.getLogger("HWR").warning("No XMLRPCServer configured")
            return

        logging.info(
            "Starting a workflow on http://%s:%d/BES", self.bes_host, self.bes_port
        )

        self.dict_parameters["initiator"] = HWR.beamline.session.endstation_name
        self.dict_parameters["sessionId"] = HWR.beamline.session.session_id
        self.dict_parameters["externalRef"] = HWR.beamline.session.get_proposal()
        try:
            self.dict_parameters["sample"] = HWR.beamline.lims.find_sample_by_sample_id(
                self.dict_parameters.get("sample_lims_id")
            )

        except (RuntimeError, AttributeError):
            logging.warning("Failed to fetch sample information for")

        try:
            self.dict_parameters["investigationId"] = (
                HWR.beamline.lims.session_manager.active_session.session_id
            )
        except (RuntimeError, AttributeError):
            logging.warning("Failed to fetch investigationId from HWR.beamline.lims")

        self.dict_parameters["token"] = (
            self.token
        )  # Deprecated in favor of mxcubeParameters
        self.dict_parameters["mxcubeParameters"] = {
            "host": socket.getfqdn(),
            "port": xml_rpc_server.port,
            "token": self.token,
        }

        start_URL = f"http://{self.bes_host}:{self.bes_port}/RUN/{self.workflow_name}"
        logging.getLogger("HWR").info("BES start URL: %r", start_URL)
        response = requests.post(start_URL, json=self.dict_parameters)
        if response.status_code == 200:
            self.state.value = "RUNNING"
            request_id = response.text
            logging.getLogger("HWR").info(
                "Workflow started, BES request id: %r", request_id
            )
            self.bes_workflow_id = request_id
        else:
            logging.getLogger("HWR").error("Workflow didn't start!")
            self.state.value = "ON"
